/*
 * This File is part of the iowj-project   
 * $Id: Iow.java,v 1.11 2005/11/19 12:31:03 Thomas Wagner Exp $
 * Copyright (C)2004 by Thomas Wagner
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.wagner_ibw.iow;

import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.Vector;

import sun.misc.Queue;

import com.codemercs.iow.IowKit;

/**
 * This is the base class of all the iow devices, based on the
 * com.codemercs.iow.IowKit JNI wrapper class to the iowkit library 
 * (iowki.dll for Windows, libiowkit.so for Linux).
 * @author Thomas Wagner
 */
public class Iow {
	
	/** 
	 * A constant for IOW40 device id.
	 * (decimal 5376)
	 */
	static final long IOW40ID = 0x1500;
	
	/**
	 * A constant for IOW24 device id.
	 * (decimal 5377)
	 */	
	static final long IOW24ID = 0x1501;
	
	/**
	 * A constant for IOW40 device name. 
	 */	
	static final String IOW40NAME = "IOW40";
	
	/**
	 * A constant for IOW24 device name.
	 */	
	static final String IOW24NAME = "IOW24";
					
	/**
	 * Count of available iow devices.
	 */
	private static long num;
	
	/**
	 * Vector of all scanned iow devices.
	 */
	private static Vector iowDevices = new Vector();
	
	/**
	 * Device handle of this iow device.
	 * (Used by subclasses IOW40 and IOW24.)
	 */
	protected long handle;
	
	/**
	 * Device identifier of this iow device.
	 * (Used by subclasses IOW40 and IOW24.)
	 */
	protected int id;
	
	/**
	 * Revision identifier of this iow device.
	 * (Used by subclasses IOW40 and IOW24.)
	 */
	protected int rev;
	
	/**
	 * Serial number of this iow device.
	 * (Used by subclasses IOW40 and IOW24.)
	 */
	protected String serial;

	/**
	 * Queue for receiving the current pin status report
	 * @since 0.9.3
	 */
	protected Queue currentPinStatus;
	
	/**
	 * Hashtable of all registered i2c devices.
	 * (Used by subclasses IOW40 and IOW24.)
	 */
	protected Hashtable i2cDevices = new Hashtable();

	/**
	 * Sole constructor. (For invocation by subclass constructors,
	 * typically implicit.)
	 *
	 */
	public Iow(){
		currentPinStatus = new Queue();
	}
		
	/**
	 * Open all connected iow devices.
	 * @param debug
	 * @deprecated
	 */	
	public void openAllDevices(boolean debug) {	
		openAllDevices();
	}
		
	/**
	 * Open all connected iow devices.
	 * @since 0.9.3
	 */
	public long openAllDevices() {
		//setDebug(debug);
		IowKit.openDevice();
		//scan for iow devices
		num    = IowKit.getNumDevs();
		for (int i = 0; i < num; i++) {
			long handle = IowKit.getDeviceHandle(i+1); //Zhlung beginnt ab 1
			int id = (int) IowKit.getProductId(handle);
			String serial = IowKit.getSerialNumber(handle);
			int rev = (int) IowKit.getRevision(handle);
			if(id == IOW40ID){
				Iow40 dev40 = new Iow40(handle, id, serial, rev);
				iowDevices.add(dev40);
			}
			else if(id == IOW24ID){
				Iow24 dev24 = new Iow24(handle, id, serial, rev);
				iowDevices.add(dev24);
			}
		}
		return num;
	}

	/**
	 * Set timeout for read operations.
	 * @param timeout Timeout value in ms.
	 */
	public void setTimeout(long timeout) {
		IowKit.setTimeout(handle,timeout);
	}
	
	/**
	 * Set timeout for write operations.
	 * @param timeout Timeout value in ms.
	 * @since 0.9.3
	 */
	public void setWriteTimeout(long timeout) {
		IowKit.setWriteTimeout(handle,timeout);
		
	}
	
	/**
	 * Returns the number of connected IOW devices.
	 * @return Number of connected IO Warrior devices.
	 * @since 0.9.3
	 */
	public long getNumDevices() {
		return num;
	}
	
	/**
	 * Returns the first iow40 device.
	 * @throws NoSuchElementException
	 */
	public Iow40 getIow40Device() throws NoSuchElementException {
		Iow40 dev = null;
		boolean found = false;
		for(int i = 0; i < num; i++) {
			IowInterface iowd = (IowInterface) iowDevices.get(i);
			if( iowd.getId() == IOW40ID) {
				dev = (Iow40) iowDevices.get(i);
				found = true;
				break;
			}
		}
		if(!found) {
			throw new NoSuchElementException(IOW40NAME);
		}
		return dev;
	}
		
	/**
	 * Returns the iow40 device related to the specified serial number.
	 * @param serial
	 * @throws NoSuchElementException
	 */
	public Iow40 getIow40Device(String serial) throws NoSuchElementException {
		Iow40 dev = null;
		boolean found = false;
		for(int i = 0; i < num; i++) {
			IowInterface iowd = (IowInterface) iowDevices.get(i);
			if( iowd.getId() == IOW40ID && iowd.getSerial().equalsIgnoreCase(serial)) {
				dev = (Iow40) iowDevices.get(i);
				found = true;
				break;
			}
		}
		if(!found) {
			throw new NoSuchElementException("No " + IOW40NAME + " device, serial=" + serial +" found!");
		}
		return dev;
	}		
	
	/**
	 * Returns the first iow24 device.
	 * @throws NoSuchElementException
	 */
	public Iow24 getIow24Device() throws NoSuchElementException {
		Iow24 dev = null;
		boolean found = false;
		for(int i = 0; i < num; i++){
			IowInterface iowd = (IowInterface) iowDevices.get(i);
			if( iowd.getId() == IOW24ID){
				dev = (Iow24) iowDevices.get(i);
				found = true;
				break;
			}
		}
		if(!found) {
			throw new NoSuchElementException(IOW24NAME);
		}
		return dev;
	}

	/**
	 * Returns the iow24 device related to the specified serial number.
 	 * @param serial
 	 * @return The desired IO Warrior 24 device.
 	 * @throws NoSuchElementException
	 */
	public Iow24 getIow24Device(String serial) throws NoSuchElementException {
		Iow24 dev = null;
		boolean found = false;
		for(int i = 0; i < num; i++){
			IowInterface iowd = (IowInterface) iowDevices.get(i);
			if( iowd.getId() == IOW24ID && iowd.getSerial().equalsIgnoreCase(serial)) {
				dev = (Iow24) iowDevices.get(i);
				found = true;
				break;
			}
		}
		if(!found) {
			throw new NoSuchElementException("No " + IOW24NAME + " device, serial=" + serial +" found!");
		}
		return dev;
	}	
		
	/**
	 * Close all the iow devices.
	 */
	public void closeAllDevices(){
		for(int i = 0; i < num; i++){
			IowInterface iowd = (IowInterface) iowDevices.get(i);
			IowKit.closeDevice(iowd.getHandle());
		}
	}
	
	/**
	 * Close this iow device
	 */
	public void close(){
		if(handle != 0) {
			IowKit.closeDevice(handle);
			handle = 0;
		}
		else {
			System.out.println("There is no open device!");
		}
	}

	/**
	 * Implementation for IowInterface method readReport.
	 * @param pipe
	 * @return Array of int.
	 */
	public int[] readReport(int pipe){
		int rbuf[] = IowKit.read(handle,pipe,8);
		
		int len = rbuf.length;
		for (int i = 0; i < len; i++) {
			rbuf[i] = rbuf[i] & 0xff; 
		}
		return rbuf;
	}
		
	/**
	 * Implementation for IowInterface method writeReport.
	 * @param pipe
	 * @param wbuf
	 * @return Count of written bytes.
	 */	
	public long writeReport(int pipe, int[] wbuf){
		return IowKit.write(handle,pipe,wbuf);
	}

	/**
	 * Gets the current pin status.
	 * @return Array of int.
	 * @since 0.9.3
	 */
	public int[] getCurrentPinStatus() throws UnsupportedOperationException {
		int wbuf[] = {0xff,0,0,0,0,0,0,0};
		
		if(rev < 0x1011) {
			throw new UnsupportedOperationException("Release " + Integer.toHexString(rev) + " does not support the current pin status report");
		}
										
		long ret = writeReport(1,wbuf);
		if(ret == 8){
			int rbuf[] = {0,0,0,0};
			try {
				int[] resbuf = (int[])currentPinStatus.dequeue();
				rbuf[0] = resbuf[1];
				rbuf[1] = resbuf[2];
				rbuf[2] = resbuf[3];
				rbuf[3] = resbuf[4];
			}
			catch (InterruptedException e) {
				e.printStackTrace();
			}
			return rbuf;
		}
		else {
			System.out.println("getCurrentPinStatus() writeReport returns != 8");
			return new int[0];
		}
	}
		
	
	/**
	 * Implementation for IowInterface method getHandle.
	 * (Used in subclasses IOW40 and IOW24 only.) 
	 * @return Returns the handle of this iow device.
	 */	
	public long getHandle(){
		return handle;
	}
	
	/**
	 * Implementation for IowInterface method getSerial.
	 * (Used in subclasses IOW40 and IOW24 only.) 
	 * @return The serial number of this iow device.
	 */
	public String getSerial(){
		return serial;
	}
	
	/**
	 * Implementation for IowInterface method getId.
	 * (Used in subclasses IOW40 and IOW24 only.) 
	 * @return Returns the device id of this device.
	 */
	public int getId(){
		return id;
	}
	
	/**
	 * Implementation for IowInterface method getRev.
	 * (Used in subclasses IOW40 and IOW24 only.) 
	 * @return Returns the revision id of this device.
	 * @since 0.9.3
	 */
	public int getRev(){
		return rev;
	}

	/**
 	 * Returns a <code>String</code> object that representing this iow device.
 	 * It contains information about device handle, id, revision and serial number.  
 	 */
	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append("Number of scanned devices: " + num + "\n");
		for(int i = 0; i < num; i++){
			IowInterface iowd = (IowInterface) iowDevices.get(i);
			sb.append("Device" + (i+1) + ": ");
			sb.append(iowd.getName());
			sb.append(",Handle[");
			sb.append(iowd.getHandle());
			sb.append("],Id[");
			sb.append(iowd.getId());
			sb.append("],Rev[");
			sb.append(Integer.toHexString(iowd.getRev()));
			sb.append("],Serial[");
			sb.append(iowd.getSerial());
			sb.append("]\n");
		}
		return sb.toString();
	}
	
	/**
	 * Closes all open devices and terminates the jvm.
	 * @since 0.9.3
	 */
	public void exit(int code) {
		closeAllDevices();
		System.exit(code);
	}
	
	/**
	 * Returns the version of the loaded iowkit library.
	 * @return Version of the iowkit.dll.
	 * @since 0.9.3
	 */
	public String getVersion(){
		return IowKit.version();
	}

}
