/*
 * This File is part of the iowj-project   
 * $Id: Iow24.java,v 1.12 2005/11/27 13:43:02 Thomas Wagner Exp $
 * Copyright (C)2004 by Thomas Wagner
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
package de.wagner_ibw.iow;

import java.util.Vector;

import com.codemercs.iow.IowKit;

/**
 * This class represents the iow24 device.
 * @author Thomas Wagner
 */
public class Iow24 extends Iow implements IowInterface, Runnable {
	
	/**
	 * Flag for all special mode functions. Each bit stands for one enabled or
	 * disbaled special mode function. See SpecialModeFunction constants.
	 */
	private int specialMode = 0;

	/**
	 * Vector of special mode implementations.
	 */
	private Vector smfImplementations = new Vector();		
	
	/**
	 * The ports.
	 */
	private IowPort[] ports = {new IowPort(0,"IIIIIIII"),
							   new IowPort(1,"IIIIIIII")};

	private boolean autonomous = false;  
	
	public Iow24(long handle, int id, String serial, int rev){
		super.handle = handle;
		super.id = id;
		super.serial = serial;
		super.rev = rev;
		new SpecialModeReadThread().start();
	}

	/**
	 * 
	 * @param impl
	 */
	public void addSpecialModeFunctionImpl(SpecialModeFunction impl)
		throws UnsupportedOperationException {
		
		if(!impl.checkCompatibility(id, rev, specialMode))
			throw new UnsupportedOperationException(Iow.IOW24NAME + ", rev " + Integer.toHexString(rev) + " does not support " + impl.getName() + " at this time!");
		
		if (!smfImplementations.contains(impl)) {
			//set related special mode bit
			specialMode = specialMode | impl.getSpecialModeFuncionId();
			//enabel special mode function
			long ret = writeReport(1, impl.getEnableReport()); //check ret!
			//set special bits in bitmask
			for (int i = 0; i < impl.getIow24SpecialBits().length; i++) {
				ports[i].setSpec(ports[i].getSpec() | impl.getIow24SpecialBits()[i]);
			}
			//set iow reference
			impl.setIowDevice(this);
			//add implementation to list
			smfImplementations.add(impl);
		}
	}

	/**
	 * @param impl
	 */
	public void removeSpecialModeFunctionImpl(SpecialModeFunction impl) {
		if (smfImplementations.contains(impl)) {
			//reset related special mode bit
			specialMode = specialMode & (255 - impl.getSpecialModeFuncionId());
			//disable special mode function		
			long ret = writeReport(1, impl.getDisableReport()); //check ret!
			//reset special bits in bitmask
			for (int i = 0; i < impl.getIow24SpecialBits().length; i++) {
				ports[i].setSpec(ports[i].getSpec() & (255 - impl.getIow24SpecialBits()[i]));
			};
			//delete iow reference
			impl.setIowDevice(null);
			//remove implementation from list
			smfImplementations.remove(impl);
		}
	}
			
	/*
	 *  (non-Javadoc)
	 * @see de.wagner_ibw.iow.IowInterface#readIOPorts()
	 * @since 0.9.3 with long return value 
	 */
	public long readIOPorts(){
		int[] rbuf = IowKit.read(handle,0,5);
		long ret = 0;
		if(rbuf.length > 0) {
			ports[0].setDataFromRead(rbuf[1] & 0xff);
			ports[1].setDataFromRead(rbuf[2] & 0xff);
			ret = rbuf[1] & 0xff;
			ret = ret + ((rbuf[2] & 0xff) << 8);
		}
		return ret;
	}

	public void readIOPortsImmediate(){
		int[] ri = IowKit.readImmediate(handle);
		if(ri.length > 0) {
			System.out.println("Iow24.read immediate()->"+ ri[0]+","+ri[1]);
			ports[0].setDataFromRead(ri[1]);
			ports[1].setDataFromRead(ri[2]);
		}
		else {
			System.out.println("Iow24.read immediate()-> no result!");		
		}
	}

	/**
	 * Gets the current pin status.
	 * @return Count of written bytes to iow device.
	 * @since 0.9.3
	 */
	public long scanPorts(){
		long ret = 0;
		int rbuf[] = getCurrentPinStatus();
		
		if(rbuf.length > 0) {
			ports[0].setDataFromRead(rbuf[0] & 0xff);
			ports[1].setDataFromRead(rbuf[1] & 0xff);
			ret = rbuf[0] & 0xff;
			ret = ret + ((rbuf[1] & 0xff) << 8);
		}
		return ret;
	}
				
	public void writeIOPorts(){
		if(ports[0].isChangeFlagSet() ||
		   ports[1].isChangeFlagSet()) {
			int[] wbuf = {0,0,0};
			wbuf[1] = ports[0].getDataToWrite();
			wbuf[2] = ports[1].getDataToWrite();
			long ret = IowKit.write(handle,0,wbuf);
			//*deb*/System.out.println("[Iow24.writeIOPorts() ret=" +ret);
			if(ret == 3) {
				ports[0].clearChangeFlag();
				ports[1].clearChangeFlag();
			}	
		}
		else {
			//*deb*/System.out.println("No writeIOPorts() needed!");
		}
	}
	
	public void run(){
		while (autonomous) {
			if (Thread.interrupted()) {
				break;
			}
			//*deb*/System.out.println("Thread: readIOPorts()...");
			readIOPorts();
			//*deb*/System.out.println("Thread: readIOPorts() done");
		}	
	}
	
	/**
	 * This method sets the autonomous mode either on or off.
	 * Autonomous mode means, that class uses readIOPorts() in a loop in a own thread.
	 * @param status true switches the mode on, false switches the mode off.
	 */
	public void autonomous(boolean status){
		Thread t;
		if(status){
			autonomous = true;	
			t = new Thread(this);
			t.start();
		}
		else {
			autonomous = false;
		}
	}
			
	public String getName(){
		return IOW24NAME;
	}
	
	public IowPort getPort(int port) throws IllegalArgumentException {
		checkPort(port);
		return ports[port];
	}
		
	/**
	 * Set the given bit froma the given port.
	 * @param port Port number (0...1).
	 * @param bit Bit number (0...7).
	 */
	public void setBit(int port, int bit)
		throws IllegalArgumentException {
		checkPort(port);
		ports[port].setBit(bit);
	}

	/**
	 * Clear the given bit froma the given port.
	 * @param port Port number (0...1).
	 * @param bit Bit number (0...7).
	 */
	public void clearBit(int port, int bit)
		throws IllegalArgumentException {
		checkPort(port);
		ports[port].clearBit(bit);
	}

	/**
	 * Set the pattern to the given port.
	 * @param port Port number (0...1).
	 * @param pattern
	 */
	public void setPort(int port, int in)
		throws IllegalArgumentException {
		checkPort(port);
		ports[port].setData(in);
	}

	/**
	 * Set the direction mask to the given port.
	 * @param port Port number (0...1).
	 * @param direction 0-output, 1-input
	 * @since 0.9.3
	 */
	public void setDirection(int port, int direction)
		throws IllegalArgumentException {
		checkPort(port);
		ports[port].setDirection(direction);
	}
		
	/**
	 * Returns a <code>String</code> object representing this iow device.
	 * It contains information about device handle, id , revision, 
	 * serial number and the status of port 0 ... port 1.  
	 */
	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append(getName());
		sb.append(":Handle[");
		sb.append(getHandle());
		sb.append("],Id[");
		sb.append(getId());
		sb.append("],Rev[");
		sb.append(Integer.toHexString(getRev()));
		sb.append("],Serial[");
		sb.append(getSerial());
		sb.append("],SpecialMode: ");
		if ((specialMode & SpecialModeFunction.SMF_LCD_ID) == SpecialModeFunction.SMF_LCD_ID)
			sb.append("LCD ");
		if ((specialMode & SpecialModeFunction.SMF_RC5_ID) == SpecialModeFunction.SMF_RC5_ID)
			sb.append("RC5 ");
		if ((specialMode & SpecialModeFunction.SMF_CPS_ID) == SpecialModeFunction.SMF_CPS_ID)
			sb.append("CPS");
		if ((specialMode & SpecialModeFunction.SMF_SPI_ID) == SpecialModeFunction.SMF_SPI_ID)
			sb.append("SPI");
		if ((specialMode & SpecialModeFunction.SMF_I2C_ID) == SpecialModeFunction.SMF_I2C_ID) {
			sb.append("I2C\n\t");
			for (int i = 0; i < smfImplementations.size(); i++) {
				SpecialModeFunction smf = (SpecialModeFunction) smfImplementations.get(i);
				if (smf.getSpecialModeFuncionId() == SpecialModeFunction.SMF_I2C_ID) {
					sb.append(smf.toString());
					break;
				}
			}
		}
		sb.append("\n\t");
		sb.append(ports[0].toString());
		sb.append("\n\t");
		sb.append(ports[1].toString());
		/*
		if(modeRC5) {
			sb.append("\n\t");
			sb.append(rc5.toString());
		}
		*/
		sb.append("\n");
		return sb.toString();
	}
	
	private void checkPort(int port) throws IllegalArgumentException {
		if(port < 0 || port > 1 )
			throw new IllegalArgumentException("Invalid port number (" + port + ")");
	}
	
	/**
	 * Read thread for special mode functions.
	 * @author Thomas Wagner
	 */
	public class SpecialModeReadThread extends Thread {
			
		/* (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			do {
				if (Thread.interrupted()) {
					break;
				}
				int[] rbuf = readReport(1);
				if( rbuf.length > 0) {
					//*deb*/System.out.println("ReadThread.readReport(1): " + Integer.toHexString(rbuf[0]));
					
					for (int i = 0; i < smfImplementations.size(); i++) {
						SpecialModeFunction smf = (SpecialModeFunction) smfImplementations.get(i);
						if (smf.matchReportId(rbuf[0])) {
							smf.reportReceived(rbuf);
							break;
						}
					}
					if (rbuf[0] == 0xff) { //current pin status report
						//*deb*/System.out.println("current pin status report received !");
						currentPinStatus.enqueue(rbuf);
					}
				}
			} while (true);
		}
	}
}
