/* 
 * This File is part of the iowj-project   
 * $Id: LM75.java,v 1.6 2005/07/14 19:09:53 Thomas Wagner Exp $
 * Copyright (C)2004 by Thomas Wagner
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package de.wagner_ibw.iow;

import sun.misc.Queue;

/**
 * This is the implemenation of the LM75 device , "Digital Temperature
 * Sensor and Thermal Watchdog with Two-Wire Interface", by National
 * Semiconductors.<br>
 * <p>This class almost supports features of the device:
 * <ul>
 *   <li>Read the current temperature.</li>
 *   <li>Setting OS and HYST temperature.</li>
 *   <li>Setting Comparator/Interrupt mode for the O.S. pin.</li>
 * 	 <li>Setting O.S. Polarity</li>
 * </ul>
 * Unsupported features are:
 * <ul>
 *   <li>Resetting the device.</li>
 *   <li>Setting the device to low power shutdown mode.</li>
 *   <li>Reporting the Fault Queue.</li>
 * </ul>
 * </p>
 * @author Thomas Wagner
 */
public class LM75 implements I2CDevice {

	/**
	 * Reference to the underlying IOW device.
	 */
	private IowInterface iow;
	
	/**
	 * Slave address of this LM75.
	 */
	private I2CAddress devAdr;
	
	/**
	 * LM75 device specific constants.
	 */
	public final static String NAME = "LM75";
	
	/**
	 * Group 1 part of slave address for LM75
	 * (LM75 device specific constant).
	 */
	public final static int CLASS = 0x09;
	
	/**
	 * Constant for register pointer for 'Temperature' register
	 * (LM75 device specific constant).
	 */
	public final static int REG_TEMP = 0;

	/**
	 * Constant for register pointer for 'Configuration' register
	 * (LM75 device specific constant).
	 */
	public final static int REG_CONFIG = 1;

	/**
	 * Constant for register pointer for 'THyst Set Point' register
	 * (LM75 device specific constant).
	 */
	public final static int REG_TEMP_HYST = 2;
	
	/**
	 * Constant for register pointer for 'TOS Set Point' register
	 * (LM75 device specific constant).
	 */
	public final static int REG_TEMP_OS = 3;

	/**
	 * Constant for 'Comparator' mode.
	 */
	public final static int MODE_COMPARATOR = 0;

	/**
	 * Constant for 'Interrupt' mode.
	 */
	public final static int MODE_INTERRUPT = 1;

	/**
	 * Constant for O.S. output polarity 'active low'.
	 */
	public final static int POLARITY_ACTIVE_LOW = 0;

	/**
	 * Constant for O.S. output polarity 'active high'.
	 */
	public final static int POLARITY_ACTIVE_HIGH = 1;
	
	/**
	 * Current setted OS temperature (initialized with factory default value).
	 */
	private double tOs = 80.0;

	/**
	 * Current setted THYST temperature (initialized with factory default value).
	 */
	private double tHyst = 75.0;
	
	/**
	 * Current temperature (measured by LM75).
	 */
	private double tCurrent = 0;

	/**
	 * Current mode (initialized with factory default value, read from LM75).
	 */
	private int mode = 0;

	/**
	 * Current polarity (initialized with factory default value, read from LM75).
	 */
	private int polarity = 0;

	/**
	 * Flag for last register access.
	 */
	private int lastRegisterPointer = -1;

	private Queue writeAckQueue;
	private Queue readQueue;
	
	/**
	 * Constructor
	 * @param iow Reference to IOW device which this LM75 is connected.
	 * @param deviceAddress Group 2 part of slave address (possible values 0...7).
	 * @throws Exception If anything goes wrong.
	 */
	public LM75(int deviceAddress) throws Exception{
		devAdr = new I2CAddress(CLASS,deviceAddress);
		writeAckQueue = new Queue();
		readQueue = new Queue();
	}
	
	
	private void init() throws Exception{
		System.out.println("LM75.init()...");
//		iow.addI2CDevice(devAdr,NAME); 	 
		//check adr im Bereich (manche chips nur ein bit
	 	
		//check adr schon vergeben (in dieser device classe, also devAdr checken?
	 	
		//set the register pointer to the temperature register
		
		getT();
		getTOs();
		getTHyst();
		System.out.println("LM75.init()end, "+this.toString());
	}
	
	
	/**
	 * Returns Current temperature measured by this LM75.
	 * @return Temperature as double value from this LM75.
	 * @throws Exception If anything goes wrong.
	 */
	public double getT()throws Exception{
		setRegisterPointer(REG_TEMP);
		tCurrent = getTemperature();
		return tCurrent;
	}
	
	
	public void getConfiguration() throws Exception {
		setRegisterPointer(REG_CONFIG);
		
		int wbuf[] = {3,1,devAdr.getReadAddress(),0,0,0,0,0};
		long ret = iow.writeReport(1,wbuf);
		//*deb*/System.out.println("[getConfiguration] write i2c "+ret+" byte(s)");		
	
		//int rbuf[] = iow.readReport(1);
		int rbuf[] = (int[]) readQueue.dequeue();
		if(rbuf.length == 0){
			throw new Exception("[getConfiguration] Read acknowledge ("+this+") failed!");
		}
		//*deb*/System.out.println("[getConfiguration] read i2c "+rbuf.length+" byte(s)");
		if((rbuf[1] & 0x80) == 0x80){
			//*deb*/System.out.println("\t[getConfiguration] read operation failed");
			throw new Exception("[getConfiguration] Read ("+this+") failed!");
		}
		else {
			//*deb*/System.out.println("\t[getConfiguration] read operation ok");
		}
		//*deb*/System.out.println("\t[getConfiguration] "+ (rbuf[1] & 0x0007) +" byte(s) confirmed");
			
		int mode = (rbuf[2] & 0x20) >> 1;
		int polarity = (rbuf[2] & 0x40) >> 2;
		//*deb*/System.out.println("Mode: "+mode+", Polarity: "+polarity);
	}
	
	
	/**
	 * Returns THYST set point stored in this LM75.
	 * @return HYST temperature as double value from this LM75.
	 * @throws Exception If anything goes wrong.
	 */
	public double getTHyst()throws Exception{
		setRegisterPointer(REG_TEMP_HYST);
		tHyst = getTemperature();
		return tHyst;
	}
	
	
	/**
	 * Returns TOS set point stored in this LM75.
	 * @return OS temperature as double value from this LM75.
	 * @throws Exception If anything goes wrong.
	 */
	public double getTOs() throws Exception {
		setRegisterPointer(REG_TEMP_OS);
		tOs = getTemperature();
		return tOs;
	}
	
	
	public void setConfiguration(int mode, int polarity) throws Exception {
		int conf = (mode << 1) | (polarity << 2);
		setRegister(REG_CONFIG,1,conf,0);
		//this.mode = mode;
		//this.polarity = polarity;
	}
	
	
	/**
	 * Sets THYST set point in this LM75.
	 * @param value Desired HYST temperature as doubel value.
	 */
	public void setTHyst(double value)throws Exception{
		LM75Temp temp = new LM75Temp(value);
		this.tHyst = temp.getTemp();
		setRegister(REG_TEMP_HYST,2,temp.getMsdb(),temp.getLsdb());
	}
	
	
	/**
	 * Sets TOS set point in this LM75.
	 * @param value Desired OS temperature as doubel value.
	 */
	public void setTOs(double value)throws Exception{
		LM75Temp temp = new LM75Temp(value);
		this.tOs = temp.getTemp();
		//*deb*/System.out.println("setTOS("+value+") called... ("+temp.getMsdb()+","+temp.getLsdb()+")");
		setRegister(REG_TEMP_OS,2,temp.getMsdb(),temp.getLsdb());
	}
	
	/**
	 * Getter for slave address of this I2C device.
	 * @return Slave address of this I2C device.
	 */
	public I2CAddress getAdr(){
		return devAdr;
	}
	

	/**
	 * Getter for mode of this LM75 device.
	 * @return Mode of this I2C device.
	 */
	public int getMode() {
		return mode;
	}
	
	
	/**
	 * Getter for polarity of this LM75 device.
	 * @return Polarity of this I2C device.
	 */
	public int getPolarity(){
		return polarity;
	}
	

	/**
 	 * Support method for setting register pointer. 
 	 * @param pointer Desired register (possible values:
 	 * REG_TEMP (0), REG_CONFIG (1), REG_TEMP_HYST (2), REG_TEMP_OS (3)).  
 	 * @throws Exception If anything goes wrong.
 	 */
	private void setRegisterPointer(int pointer)throws Exception{
		if( pointer == lastRegisterPointer) return; //nothing to do
		
		int rbuf[] = null;
	  	int wbuf[] = {2,0xc2,devAdr.getWriteAddress(),pointer,0,0,0,0};
	  	long ret = iow.writeReport(1,wbuf);
	  	//*deb*/System.out.println("[setRegisterPointer] write i2c "+ret+" byte(s)");		
		
	  	//rbuf = iow.readReport(1);
		rbuf = (int[]) writeAckQueue.dequeue();
		if(rbuf.length == 0){
			throw new Exception("[setRegisterPointer] Read acknowledge ("+this+") failed!");
	  	}
	  	//*deb*/System.out.println("[setRegisterPointer] read i2c "+rbuf.length+" byte(s)");
	 	int flags = rbuf[1];
	  	if((flags & 0x80) == 0x80){
			//*deb*/System.out.println("\t[setRegisterPointer] read operation failed");
			throw new Exception("[setRegisterPointer] Read ("+this+") failed!");
	  	}
	  	else {
		  	//*deb*/System.out.println("\t[setRegisterPointer] read operation ok");
		  	lastRegisterPointer = pointer;
		}
		//*deb*/System.out.println("\t[setRegisterPointer] "+ (flags & 0x0007) +" byte(s) read");
		
	}
	
	
	/**
	 * Support method for getting temperature value for register.
	 * specified by preseted register pointer.
	 * @return Temperature as double value.
	 * @throws Exception If anything goes wrong.
	 */
	private double getTemperature() throws Exception{
		int wbuf[] = {3,2,devAdr.getReadAddress(),0,0,0,0,0};
		long ret = iow.writeReport(1,wbuf);
		//*deb*/System.out.println("[getTemperature] write i2c "+ret+" byte(s)");		
	
		//int rbuf[] = iow.readReport(1);
		int rbuf[] = (int[]) readQueue.dequeue();
		if(rbuf.length == 0){
			throw new Exception("[getTemperature] Read acknowledge ("+this+") failed!");
		}
		//*deb*/System.out.println("[getTemperature] read i2c "+rbuf.length+" byte(s)");
		if((rbuf[1] & 0x80) == 0x80){
			//*deb*/System.out.println("\t[getTemperature] read operation failed");
			throw new Exception("[getTemperature] Read ("+this+") failed!");
		}
		else {
			//*deb*/System.out.println("\t[getTemperature] read operation ok");
		}
		//*deb*/System.out.println("\t[getTemperature] "+ (flags & 0x0007) +" byte(s) confirmed");
				
		LM75Temp temp = new LM75Temp(rbuf[2],rbuf[3]);
		return temp.getTemp();
	}
	
	
	/**
	 * Support method for setting writeable registers. Possibly registers are
	 * CONFIG, TOS and THYST.
	 * @param regAdr Desired register (possible values: REG_CONFIG (1), REG_TEMP_HYST (2), REG_TEMP_OS (3)).
	 * @param count Counter of bytes of pay load (possible values: 1 or 2).
	 * @param firstByte First byte of the pay load (must set).
	 * @param secondByte Second byte of the pay load (if counter is 1 this param must be set 0).
	 * @throws Exception If anything goes wrong.
	 */
	private void setRegister(int regAdr, int count, int firstByte, int secondByte)throws Exception{
		int rbuf[] = null;
		int wflags = 0xc0;
		wflags = wflags | count +2;
		int wbuf[] = {2,wflags,devAdr.getWriteAddress(),regAdr,firstByte,secondByte,0,0};
		long ret = iow.writeReport(1,wbuf);
		//*deb*/System.out.println("[setRegister] write i2c "+ret+" byte(s)");		
		lastRegisterPointer = regAdr;
		
		//rbuf = iow.readReport(1);
		rbuf = (int[]) writeAckQueue.dequeue();
		if(rbuf.length == 0){
			throw new Exception("[setRegister] Read acknowledge ("+this+") failed!");
		}
		//*deb*/System.out.println("[setRegister] read i2c "+rbuf.length+" byte(s)");
		int flags = rbuf[1];
		if((flags & 0x80) == 0x80){
			//*deb*/System.out.println("\t[setRegister] read operation failed");
			throw new Exception("[setRegister] Read ("+this+") failed!");
		}
		else {
			//*deb*/System.out.println("\t[setRegister] read operation ok");
		}
		//*deb*/System.out.println("\t[setRegister] "+ (flags & 0x0007) +" byte(s) confirmed");
		
	}
	
	
	/**
	 * Returns Information about this LM75 as String.
	 * @return String representation of information obout this LM75
	 * (slave address, temperatures, ...).
	 */
	public String toString(){
		StringBuffer sb = new StringBuffer();
		
		sb.append("LM75:");
		sb.append("SlaveAddress[");
		sb.append(devAdr);
		sb.append("],tCurr[");
		sb.append(tCurrent);
		sb.append("],tOs[");
		sb.append(tOs);
		sb.append("],tHyst[");
		sb.append(tHyst);
		sb.append("]");
		
		return sb.toString();
	}
	
	
	/**
	 * Inner class for LM75 temperature representation.
	 */
	private class LM75Temp{
		
		/**
		 * Constant for minimum temperature valid for LM75.
		 */ 
		final static double TEMP_MIN = -55.0;

		/**
		 * Constant for maximum temperature valid for LM75.
		 */ 
		final static double TEMP_MAX = 125.0;
			
		/**
		 * Most significant data byte of this temperature.
		 */ 
		int msdb;
	
		/**
		 * Most significant data byte of this temperature.
		 */
		int lsdb;
		
		/**
		 * Double representation of this temperature.
		 */ 
		double temp;
		
		
		/**
		 * Constructor for temperature seperated in msdb (most significant data byte)
		 * and lsdb (least significant data byte).
		 * @param msdb Most significant data byte read from LM75.
		 * @param lsdb Least significant data byte read from LM75.
		 */
		public LM75Temp(int msdb, int lsdb){
			this.msdb = (byte) msdb; //save the sign!
			this.lsdb = lsdb & 0x80;
			temp = (byte) msdb; //save the sign!
			if((lsdb & 0x80) == 0x80){
				if((msdb & 0x80) == 0x80) temp = temp - 0.5;
				else temp = temp + 0.5;
			}
		}

		
		/**
		 * Constructor for temperature as double value.
		 * @param temp Temperature as double value.
		 */
		public LM75Temp(double temp){
			if(temp > TEMP_MAX) temp = TEMP_MAX;
			if(temp < TEMP_MIN) temp = TEMP_MIN;
			
			this.temp = temp;
			 
			double fest;
			double rest;
				
			if( temp >= 0){
				fest = Math.floor(temp);
				rest = temp - fest;
				msdb = (int)fest;
				if(rest == 0) lsdb = 0x0;
				else if(rest < 0.25) lsdb = 0x0;
				else if(rest >= 0.26 && rest < 0.75) lsdb = 0x80;
				else { lsdb = 0; msdb ++; }
			}
			else {
				fest = Math.ceil(temp);
				rest = temp - fest;
				msdb = (int)fest;
				if(rest == 0) lsdb = 0x0;
				else if(rest > -0.25) lsdb = 0x0;
				else if(rest <= -0.26 && rest > -0.75) lsdb = 0x80;
				else { lsdb = 0; msdb --;}
			}
		}

		
		/**
		 * Getter for msdb.
		 */ 
		int getMsdb(){
			return msdb;
		}
		
		
		/**
		 * Geter for lsdb.
		 */
		int getLsdb(){
			return lsdb;
		}

		
		/**
		 * Getter for double value.
		 */
		double getTemp(){
			return temp;
		}
	}
	/* (non-Javadoc)
	 * @see de.wagner_ibw.iow.I2CDevice#getI2cAddress()
	 */
	public int getI2cAddress() {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see de.wagner_ibw.iow.I2CDevice#getName()
	 */
	public String getName() {
		return NAME;
	}

	/* (non-Javadoc)
	 * @see de.wagner_ibw.iow.I2CDevice#getI2cClass()
	 */
	public int getI2cClass() {
		return CLASS;
	}

	/* (non-Javadoc)
	 * @see de.wagner_ibw.iow.I2CDevice#reportReceived(int[])
	 */
	public void reportReceived(int[] readBuffer) {
		/*deb*/System.out.println("LM75.reportReceived() received report: " + readBuffer[0]);
		
		if(readBuffer[0] == 0x02)
			writeAckQueue.enqueue(readBuffer);
		else
			readQueue.enqueue(readBuffer);
			
	}

	/* (non-Javadoc)
	 * @see de.wagner_ibw.iow.I2CDevice#setIowDevice(de.wagner_ibw.iow.IowInterface)
	 */
	public void setIowDevice(IowInterface iow) {
		this.iow  = iow;
		if (iow != null) {
			try {
				init();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

}
